# Common definitions for Pathena application
# (used by client, server and all utilities)

import os, sys, time, string, shutil, re
from operator import add, and_
from StringIO import StringIO
from traceback import print_exc
#import ConfigParser

from app_config import *

import exceptions
class StartUpError(exceptions.Exception):  pass
class UserError(exceptions.Exception):  pass
class FalseStart(exceptions.Exception): pass
class OffNominal(exceptions.Exception): pass

class CommandFailure(exceptions.Exception): pass
class CommandTimeout(exceptions.Exception): pass
class CommandAbort(exceptions.Exception): pass

class ConnectionFailure(exceptions.Exception):  pass
class DatabaseError(exceptions.Exception):  pass

class SearchTermError(exceptions.Exception):  pass


# ---------- Global constants ---------- #

indexing_log_name  = 'indexing_log'
indexing_err_log   = 'index_err_log'
pg_server_log_name = 'pg_server'
db_ops_log_name    = 'ops_log'

pathena_dir = user_directory = os.path.expandvars('$PATHENA')

pathena_subdirs = ('bin', 'config', 'data', 'lib', 'log',
                   'profiles', 'tmp', 'types')  ###

bin_dir, config_dir, data_dir, lib_dir, log_dir, \
         profile_dir, tmp_dir, types_dir = \
    [ os.path.join(pathena_dir, d)+os.sep for d in pathena_subdirs ]

identity_fn     = lambda arg: arg
null_proc       = lambda *args: None
null_string_fn  = lambda *args: ''

any_file_type   = '.*'

spawn_bin = os.path.join(bin_dir, 'spawn_with_timeout')


# Implicit file types will always be indexed.

implicit_content_types = ('node/direc', 'node/link', 'exec/script')
#                         'exec/binary',  # binary is path only

implicit_file_types = ('node/direc', 'node/link', 'exec/binary', 'exec/script')


# ---------- Configuration classes ---------- #

from ConfigParser import ConfigParser

# Introduce a slightly enhanced ConfigParser that allows comments at
# ends of lines.  Strip comments before applying 'get' methods.

class ConfigParserCommentless(ConfigParser):

    def __init__(self, *args):
        ConfigParser.__init__(self, *args)

    def get(self, *args, **kargs):
        value = ConfigParser.get(self, *args, **kargs)
        return value.split('#')[0].strip()

    def getint(self, *args, **kargs):
        return int(self.get(*args, **kargs))


# ---------- Utility procedures ---------- #

_temp_file_template = os.path.join(tmp_dir, '%s_%%s' % os.getpid())
_temp_index = 0

def new_temp_file_name():
    global _temp_index
    _temp_index += 1
    return _temp_file_template % _temp_index

def quote_quote(text):
    return text.replace("'", "''")

def unique_list(seq):
    result = []
    for item in seq:
        if item not in result: result.append(item)
    return result

def two_digit_min(value, divisor):
    scaled = value / divisor
    if scaled < 10.0:
        return str(round(scaled, 1))
    else:
        return str(int(round(scaled)))

def approx_size(size, align=0):
    if size < 1000:
        if align:
            return '%d ' % size
        else:
            return '%d' % size
    elif size < 999500:
        return '%sK' % two_digit_min(size, 1000.0)
    elif size < 999500000:
        return '%sM' % two_digit_min(size, 1000000.0)
    else:
        return '%sG' % two_digit_min(size, 1000000000.0)


# Convert number of seconds into string approximation (e.g., "3.2 days").

def approx_duration(duration):
    # thresholds are 0.95 * sec_per_unit
    if duration > 29979079:
        return '%s years' % two_digit_min(duration, 31556926.0)
    elif duration > 82080:
        return '%s days' % two_digit_min(duration, 86400.0)
    elif duration > 3420:
        return '%s hours' % two_digit_min(duration, 3600.0)
    else:
        return '%s minutes' % two_digit_min(duration, 60.0)

    
# Find all positions where capitalizations occur within a string and
# insert a blank before the capital letter.  Any transition from a
# non-upper-case letter to an upper-case letter is considered a
# capitalization site.  Example: 'getNextItem' ==> 'get Next Item'

def separate_capitalized(s):
    return ''.join([ ' '[ : int(prev.isupper() < next.isupper())] + next
                     for prev, next in zip('A' + s[:-1], s) ])

#def capital_split(s):
#    return [ ' '[ : int(prev.isupper() < next.isupper())] + next
#                     for prev, next in zip('A' + s[:-1], s) ]

# Pattern for splitting a string into alphanumeric substrings.
# '_' excluded from alphanumeric char set.

alphanum_split_pattern = re.compile('[\W_]+', re.U)


regexp_metachars = '*+?(|)-^$.!:{}[],='

# Check if (unescaped) regexp metachars appear in a string.

def has_regexp_metachars(str):
    prev_c = ' '
    for c in str:
        if c in regexp_metachars and prev_c != '\\': return 1
        else: prev_c = c
    return 0

# Escape all regexp metachars in a string.

def regexp_escape(str):
    result = str
    for c in unique_list(str):
        if c in regexp_metachars:
            result = result.replace(c, r'\%s' % c)
    return result


# Wrap string into text lines of max width, inserting newlines as needed.
# Optionally return a list of lines without newlines.
# Newlines within the string are overridden by wrapping.

def wrap_string(text, width=70, return_list=0):
    words = text.split()
    result = []
    line, total = [], 0
    for w in words:
        if total + len(w) >= width:
            if line: result.append(' '.join(line))
            line, total = [w], len(w)
        else:
            line.append(w)
            total += len(w) + 1
    if line: result.append(' '.join(line))
    if return_list: return result
    else:           return '\n'.join(result)
    

# Limit runs of blank lines to length one.

def purge_extra_blank_lines(text):
    lines = text.splitlines()
    prev_line = ''
    result = []
    for line in lines:
        if line == prev_line == '': continue
        result.append(line)
        prev_line = line
    result.append('\n')
    return '\n'.join(result)


def read_text_file(path_name, as_lines=0, max_size=-1):
    try:
        file = open(path_name)
        if as_lines: contents = file.readlines(max_size)
        else:        contents = file.read(max_size)
        file.close()
    except:            # file permission or other error
        contents = ''  ### add logging???
        try:
            file.close()
        except:
            pass
    return contents

# If supplied, error_fn accepts two string arguments, a title/category and an
# error message.  If not supplied, error messages are directed to sys.stderr.

def write_text_file(content, path_name, as_lines=0, error_fn=None):
    try:
        file = open(path_name, 'w')
        if as_lines: file.writelines(content)
        else:        file.write(content)
        file.close()
    except Exception, exc_obj:
        msg = 'Exception: %s' % ' ; '.join(map(str, exc_obj.args))
        if error_fn: error_fn('Error on File Operation', msg)
        else:        print >> sys.stderr, msg 
        try: file.close()
        except: pass


# Compound words or other "words" containing nonalphanumeric characters
# are encoded as alphanumeric strings before submission to tsearch2.
# Each such char is replaced by a string of the following forms,
# where nn... is the hexadecimal representation of the character's code.
#  'qknn'      if          ord(c) <  256
#  'qxnnn'     if   256 <= ord(c) < 4096
#  'qznnnn'    if  4096 <= ord(c)

def encode_char_alnum(c):
    if c.isalnum(): return c
    code = ord(c)
    if   code < 256:    return 'qk%x' % code
    elif code < 4096:   return 'qx%x' % code
    else:               return 'qz%x' % code

def encode_compound(compound):
    return ''.join(map(encode_char_alnum, compound))

_encoding_patterns = \
    (re.compile('qk..'), re.compile('qx...'), re.compile('qz....'))

def restore_special(match):
    code_str = match.group(0)[2:]
    try:
        sub = unichr(int(code_str, 16))
    except:
        sub = r'\x%s' % code_str
    return sub

def restore_unencoded_text(word):
    for patt in _encoding_patterns:
        word = patt.sub(restore_special, word)
    return word


# ----------- Utilities for creating Tsearch queries ------------

# Map search terms into form accepted by Tsearch2.  Terms are ANDed
# and '-' negations are converted to '!' form.
# Any nonalphanumeric chars are mapped into encoded form. ('qknn').

def translate_for_tsearch(term):
    if term[0] in '-!': start = 1
    else:               start = 0
    if term[start:].isalnum(): return term
    iterm = []
    for c in term:
        iterm.append(encode_char_alnum(c))
    return term[:start] + ''.join(iterm)

def Tsearch_terms(terms):
    for i in range(len(terms)):
        if terms[i][0] == '-': terms[i] = '!' + terms[i][1:]
    return '&'.join(map(translate_for_tsearch, terms))

def strip_neg(term):
    if term[0] in '-!': return term[1:]
    else:               return term


def split_and_pad_terms(query_string):
    terms = query_string.split(';')
    return [ t.strip() for t in terms ] + [''] * (4 - len(terms))
