# Database client connections are managed by this module.
# Also used by indexing functions.
# Loading this module establishes a connection to the PostgreSQL server
# for the Pathena database.  Several procedures are provided for invoking
# queries and transactions.

import os, sys, time, string, re

from StringIO import StringIO
from traceback import print_exc

from app_globals       import *
from logging_services  import *

if explain_queries:
    explain_log = open(os.path.join(log_dir, 'query_explanations'), 'a')
else:
    explain_log = None

max_tries = 4          # Attempts for connection opening


#-----------------------------------------

user_name = os.environ.get('LOGNAME') or os.environ.get('USER')

# Currently using PyGreSQL as the database adapter (client interface to
# PostgreSQL) because it is usually pre-installed under Linux.
# Another possibility is psycopg, which would need to be retrieved and
# installed first.

import pgdb       # DB-API 2.0 compliant module

def connect_to_db():
    return pgdb.connect(host='localhost:%s' % os.environ['PGPORT'],
                        user=user_name, database='pathena')

_server_cmd = '%s >> %s 2>&1 &' % (os.path.join(bin_dir, 'start_server'),
                                   os.path.join(log_dir, db_ops_log_name))

def open_db_connection():
    global _db, _cursor
    for i in range(max_tries):
        time.sleep(i)
        try:
            _db = connect_to_db()
            break                   # successful connection
        except:
            # assume failure to connect is due to server being down
            pass
        if os.path.exists(os.path.join(data_dir, 'postmaster.pid')):
            # postmaster seems to exist or is starting; wait a little longer
            continue
        elif i < max_tries - 1:
            # try to (re)start server process
            os.system(_server_cmd)
    else:
        raise ConnectionFailure     # retries exhausted

    _cursor = _db.cursor()          # Retain only a single cursor

    # Need to initialize the Python stored-procedure environment in each
    # PostgreSQL backend process.
    do_query_effect("select init_plpythonu_setup(%s, %s)",
                    os.environ['PATHENA'], str(os.getpid()))

def close_db_connection():
    _db.close()

_reset_connection = 1   # default: close and reopen after database error

def retain_db_connection():
    global _reset_connection
    _reset_connection = 0
    
    
#----------------
# Alternative database adapter is psycopg:
#
#import psycopg       # DB-API 2.0 compliant module

#connect_string = "dbname=pathena port=%s user=%s" % \
#                 (os.environ['PGPORT'], user=user_name)
#_db = psycopg.connect(connect_string)
#----------------


# NOTE: If a query contains '%' chars for LIKE operators, don't use the
# execute(str, params) form or it will give errors.

def do_query(str, *params):
    return do_query_any(str, 1, *params)

def do_query_effect(str, *params):
    do_query_any(str, 0, *params)

def do_query_any(str, use_result, *params):
    global _cursor
    try:
        if explain_log and str.lstrip().lower().startswith('select'):
            if params: _cursor.execute('explain analyze ' + str, params)
            else:      _cursor.execute('explain analyze ' + str)
            expl_lines = reduce(add, _cursor.fetchall())
            expl = reduce(add, [ '%s\n' % e for e in expl_lines ])
            emit_log_entry(explain_log,
                           'Query: %s %s\n\n%s' % (str, params, expl))
#        start = time.time()
        if params: _cursor.execute(str, params)
        else:      _cursor.execute(str)
        if use_result: return _cursor.fetchall()
        else:          return None
#        emit_log_entry(ops_log_file,
#                       'do_query: %s ...' % str[:30], start, time.time())
    except pgdb.Error, err:
        trace = StringIO()
        print_exc(1, trace)
        if _reset_connection: close_db_connection()
        trace_text = trace.getvalue()
        emit_log_entry(ops_log_file,
                       "do_query exception '%s' from pgdb on query\n  '%s'"
                       % (err, (str, params)))
        emit_log_entry(ops_log_file, trace_text)
        if _reset_connection: open_db_connection()
        raise DatabaseError

# Each command has form (query, arg1, ...)

def do_transaction(*commands):
    global _cursor
    try:
        for cmd in commands:
            rest = cmd[1:]
            if rest: _cursor.execute(cmd[0], rest)
            else:    _cursor.execute(cmd[0])
        result = _cursor.fetchall()
        _db.commit()
        return result
    except pgdb.Error, err:
        trace = StringIO()
        print_exc(1, trace)
        _db.rollback()
        if _reset_connection: close_db_connection()
        trace_text = trace.getvalue()
        emit_log_entry(ops_log_file,
                       "do_transaction exception '%s' from pgdb on commands:\n"
                       "   '%s'" % (err, commands))
        emit_log_entry(ops_log_file, trace_text)
        if _reset_connection: open_db_connection()
        raise DatabaseError


# Database access utilities

def getDatabaseSize(mb=0):
    size = do_query("select database_size('pathena')")[0][0]
    if mb: return size / 1000000.0
    else:  return size

def getServerStatus():
    pid_file = open(os.path.join(data_dir + 'postmaster.pid'))
    pid = pid_file.readline().strip()
    pid_file.close
    port_id = os.path.expandvars('$PGPORT')
    try:
        total, content, cached = do_query("select * from summary_stats")[0]
    except DatabaseError:
        return None
    ### add above? : where content vector not null
    db_size =  '%8.2f' % getDatabaseSize(mb=1)
    pipe = os.popen(os.path.expandvars('du -s $PATHENA/data'))
    cluster_size = '%8.2f' % (int(pipe.read().split()[0]) * 0.001024)
    pipe.close()
    return pid, port_id, db_size, cluster_size, total, content, cached


# Establish a connection for database services.

open_db_connection()


# ----------- Utilities for creating Tsearch queries ------------

def lexize_term(term):
    ### generalize from English stemmer to other dictionaries
    return do_query("select lexize('en_stem', %s)", term)[0][0][1:-1]

# Terms are vacuous if all lexemes match stop words.

def vacuous_content_query(terms):
    # will be [] if all terms are stopwords:
    lexized_terms = [ lexize_term(strip_neg(t)) for t in terms ]
    return not filter(None, lexized_terms)    # remove null lexemes

# Terms are vacuous if all numeric.

def vacuous_path_query(terms):
    term_checks = [ not strip_neg(t).isdigit() for t in terms ]
    return not filter(None, term_checks)    # remove numeric cases
