# Basic operations for file encoding are included in this module.
# Also included are class definitions for predefined file types.
# Modules in ~/.pathena/types/ will subclass these definitions.


import os, re
from StringIO import StringIO

from app_globals       import *
from logging_services  import *
from platform_utilities import *

user_path_prefix = os.path.expanduser('~/')
user_path_prefix_len = len(user_path_prefix)


def extract_plain_text(path_name, as_lines=0):
    return read_text_file(path_name, as_lines, max_indexed_len)


# Use this shell command when primary conversion commands are unavailable:

if check_os_command('strings'):
    backup_conversion = None
else:
    backup_conversion = make_shell_command('strings %s')


# Pattern for splitting a string into alphanumeric substrings.
# '_' excluded from alphanumeric char set.

alphanum_split_pattern = re.compile('[\W_]+', re.U)

# Attributes of indexed file types and others.  This module will be
# imported by all file type handlers in subdirectory "types".

indexing_attribs = (
    'text',         # full-text indexing applied to entire file
    'excerpt',      # text form, only excerpt(s) indexed
    'encoded',      # encoded form, output of external converter indexed
    'path_list',    # directory contents indexed (table of contents, etc)
    'binary',       # nontext, unindexed binary content
    'other',        # file contents unknown and unindexed
    'oversize',     # content would be indexed if file were shorter
    'special',      # special conventions for indexing
    )
### need a 'source' attrib ?

# implicit_types = ('other', 'oversize', 'direc', 'link', 'exec', 'script')


def missing_converter(name):
    # 
    def dummy_converter(path=None):
        return '''The preferred converter for this file type, %s,
is unavailable.  Consider installing '%s' if indexing and
viewing of extracted text for this file type is desirable.''' % (name, name)
    return dummy_converter

# Conversion provided by external utility or OS command if it exists.
# Command arg should be either a string containing %s where the path
# needs to be substituted, or a callable function that makes use of
# the external utility.

def make_converter_command(name, command):
    converter_exists = check_os_command(name) == None
    if not converter_exists and not backup_conversion:
        def file_converter(path):
            return missing_converter(name)()
    elif not converter_exists:
        file_converter = backup_conversion
    elif callable(command):
        file_converter = command
    else:
        file_converter = make_shell_command(command)
    return converter_exists or backup_conversion, file_converter


##########

# Encode a string by encoding its compound words.  A compound word is
# defined as a character sequence containing either alphanumeric or
# selected special characters.  The component words of compounds (those
# alphanumeric strings separated by either special characters or capital
# letter transitions) are collected and returned as a separate string.
# Underscores are implicit special characters.  Since this will be used
# for indexing, only the "wordy" portions of source text are retained.

# A crude encoding is used that maps each non-alphanumeric char into a
# string such as 'qknn', where nn is the text form of the char code.
# (See encode_compound for full story.)  This encoding is unlikely to
# clash with real strings.  Support for custom compound words in
# tsearch2 would be much preferred, of course.

# Optionally discard common "noise" words such as strings of the
# form 0x00abcd and single-character words.

def encode_for_indexing(original, specials, keep_all=1):
    # if specials = 1, treat entire original string as a compound word
    if specials == 1:
        orig_compounds = [original]
    else:
        # certain specials need protection before substitution in patterns
        esc_specials = specials.replace('-','\-').replace(']','\]')
        pattern = re.compile(r'[^\w%s]+' % esc_specials)
        # pattern = re.compile(r'([^\w%s]*)([\w%s]+)'
        # % (esc_specials, esc_specials))
        orig_compounds = pattern.split(original)
    
    full_words, subwords = [], []
    for compound in filter(None, orig_compounds):
        if not(compound.isalnum() and compound.islower()):
            ws = alphanum_split_pattern.split(separate_capitalized(compound))
            if len(ws) > 1:
                if keep_all:
                    subwords.extend(ws)
                else:
                    for w in ws:
                        if len(w) > 1 and not w.startswith('0x'):
                            subwords.append(w)
                compound = encode_compound(compound)
        if keep_all or len(compound) > 1 and not compound.startswith('0x'):
            full_words.append(compound)
    return full_words, subwords

# For source files, identifiers containing nonalphanumeric characters
# are recognized as compound words.  Also included are single-word
# identifiers having a special-char prefix or suffix.  These compounds
# are parsed and replaced by their encoded forms, then their subwords are
# added to the end of each line before submitting the file for indexing.
# The special characters to be encoded are specific to the source file type.  

def extract_source_text(path_name, id_chars):
    try:
        file = open(path_name)
        content_lines = file.readlines()
        file.close()
    except:            # file permission or other error
        try:
            file.close()
        except:
            pass
        return ''      ### add logging???
    augmented = StringIO()
    for orig_line in content_lines:
        compounds, subwords = encode_for_indexing(orig_line, id_chars, 0)
        augmented.write(' '.join(compounds))
        augmented.write(' ')
        augmented.write(' '.join(subwords))
        augmented.write('\n')
    return augmented.getvalue()


# Split an absolute path name into "words" by first breaking path into
# directory components, then further splitting into alphnumeric substrings
# and breaking up capitalized segments.  Split off file extensions as a
# special case.  Filter out purely numeric words.

def path_word_split(abs_path):
    basename = os.path.basename(abs_path)
    file, ext = os.path.splitext(basename)
    names = filter(None, abs_path.split(os.sep)[:-1] + [basename, file, ext])
    names = filter(lambda s: not s.isdigit(), names)
    full_words, subwords = [], []
    for words, subs in [ encode_for_indexing(n, 1) for n in names ]:
        full_words.extend(words)
        subwords.extend(subs)
    return unique_list(full_words + subwords)

# Join path words into a blank-delimited string.

def path_words(abs_path):
    if abs_path.startswith(user_path_prefix):
        abs_path = abs_path[user_path_prefix_len:]
    words = ' '.join(path_word_split(abs_path))
    return words.replace("'", r"\'")


# After converting a file to text, optional post-processing is performed.
### consider making methods

def post_conversion(text, attrib):
    if attrib == 'path_list':
        return ' '.join([ path_words(path) for path in text.splitlines() ])
    else:
        return text


def protected_convert_file(file_converter, path, usage=0):
    try:
        return file_converter(path)
    except CommandTimeout, msg:
        if usage in (1, 3):
            raise CommandTimeout, msg    # indexing mode
        else:
            return '***** Conversion command timed out.'
    except CommandFailure, code:
        if usage in (1, 3):
            raise CommandFailure, code    # indexing mode
        else:
            return '***** Conversion command failed with error code %s.' % code


##########

# Base class for deriving specific file type classes.
# File type classes are organized by genus and species.
# A species class inherits from a genus class to specialize its type.
# The base class is for defining genus classes only.
# Some file types may by genus-only.

class File_Base(object):

    def __init__(self, genus_name, attrib):
        self.genus_name = genus_name
        self.file_indexing_attrib = attrib

    species_name = ''      # species subclasses will override

    phylum = 'regular'     # or 'special', 'implicit'

    file_extensions = ()
    
    def file_type_name(self):
        return self.genus_name + '/' + self.species_name
    
    def file_type_pattern(self):
        if self.species_name: return self.file_type_name()
        else:                 return self.genus_name + '/.*'

    font_class = 'sans'    # or 'serif', 'fixed'

    identifier_specials = ''              # special chars (for source genus)
    
    external_viewers = None
    external_editors = None

    # List of OS/shell commands needed by file conversion methods.

    commands_needed = ()
    
    # Decide whether file conversions should be cached in database.
    # Decision is based on unconverted file sizes.  If path==None,
    # return whether conversion caching is possible at all.  Otherwise,
    # return 0 for no caching, 1 for a single cached string that can be
    # transformed as needed, and 2 for caching of display form only.
    
    def cache_conversion(self, path=None):
        return 0
    
    # This is the primary method for retrieving a file's contents and
    # converting to text, with possible variation for different uses.
    # Usage key:  0 = display, 1 = indexing, 2 = headline generation,
    #             3 = cached conversion
    
    def convert_file(self, path, usage=0):
        return extract_plain_text(path)   # might fail for binary files
    
    # If this file class uses content caching, following method is
    # needed to convert from cached form to needed variant (by usage).
    # Usage key:  0 = display, 1 = indexing, 2 = headline generation,
    
    def convert_cached_content(self, content, usage=0):
        return content

    # If a meaningful title or displayable metadata can be derived for
    # a file, override these methods.  Optional arg orig_path is used
    # when path refers to a temporary file.
    
    def extract_title(self, path, orig_path=None):
        return os.path.basename(orig_path or path)
    
    def extract_metadata(self, path, orig_path=None):
        return ''

    # Following used for special files only.  It should return a pair
    # of booleans (use_path, include_content), where include_content 
    # indicates if a file is to be indexed as part of the special class.
    # use_path indicates whether to use or suppress path name indexing.

    def recognize_file(self, path, file_name):
        return (None, None)

    # Choose what portions of a path to include for indexing.
    # Usually only needed by special file types.
    
    def path_excerpt(self, path):
        return path
    

# ---------------- Type classes for file genera ------------------

class Text_Genus(File_Base):

    def __init__(self, species_name):
        File_Base.__init__(self, 'text', 'text')
        self.species_name = species_name
    
    external_editors = text_editors

class Node_Genus(File_Base):

    def __init__(self, species_name):
        File_Base.__init__(self, 'node', 'path_list')
        self.species_name = species_name

    phylum = 'implicit'
    font_class = 'fixed'

class Exec_Genus(File_Base):

    def __init__(self, species_name, attrib='other'):
        File_Base.__init__(self, 'exec', attrib)
        self.species_name = species_name

    phylum = 'implicit'

class Other_Genus(File_Base):

    def __init__(self, species_name, attrib='other'):
        File_Base.__init__(self, 'other', 'other')
        self.species_name = species_name
        self.file_indexing_attrib = attrib

    phylum = 'implicit'
    

class Source_Genus(File_Base):

    def __init__(self, species_name, specials=''):
#        File_Base.__init__(self, 'source', 'source')
        File_Base.__init__(self, 'source', 'text')
        self.species_name = species_name

    font_class = 'fixed'
    
    external_editors = text_editors

    def convert_file(self, path, usage=0):
        if usage == 1:   # indexing
            return extract_source_text(path, self.identifier_specials)
        else:
            return extract_plain_text(path)

class Document_Genus(File_Base):

    def __init__(self, species_name):
        File_Base.__init__(self, 'doc', 'encoded')
        self.species_name = species_name

    font_class = 'serif'

    def cache_conversion(self, path=None):      # default: always cache
        return 1


class Markup_Genus(File_Base):

    def __init__(self, species_name):
        File_Base.__init__(self, 'markup', 'text')
        self.species_name = species_name

    font_class = 'fixed'

    external_editors = text_editors

class Archive_Genus(File_Base):

    def __init__(self, species_name):
        File_Base.__init__(self, 'archive', 'path_list')
        self.species_name = species_name

    font_class = 'fixed'

    def cache_conversion(self, path=None):      # default: always cache
        return 1

class Email_Genus(File_Base):

    def __init__(self, species_name):
        File_Base.__init__(self, 'email', 'excerpt')
        self.species_name = species_name

    def cache_conversion(self, path=None):      # only cache large files
        return int(path == None or os.path.getsize(path) > 100000)


# ---------- Type classes for built-in species or subgenera -----------

class Text_Type(Text_Genus):

    def __init__(self, name, *extensions):
        Text_Genus.__init__(self, name)
        self.file_extensions = extensions


class Directory_Type(Node_Genus):

    def __init__(self, name):
        Node_Genus.__init__(self, name)

    def convert_file(self, path, usage=0):
        if usage:    # indexing and headlines
            return extract_directory_children_short(path)
        else:
            return extract_directory_children_long(path)

class Link_Type(Node_Genus):

    def __init__(self, name):
        Node_Genus.__init__(self, name)

    def convert_file(self, path, usage=0):
        if usage:    # indexing and headlines
            return extract_link_target(path)
        else:
            return extract_directory_children_long(path)


class Binary_Exec_Type(Exec_Genus):

    def __init__(self, name):
        Exec_Genus.__init__(self, name, 'binary')

    def convert_file(self, path, usage=0):
        return ''
    
    def extract_metadata(self, path, orig_path=None):
        return "\nThis is a binary executable file.\n\n" \
               "Attributes according to 'file' command:\n\n    %s\n" \
               % look_up_file_attribs(path)

class Script_Exec_Type(Exec_Genus):

    def __init__(self, name):
        Exec_Genus.__init__(self, name, 'text')

    # give scripts same attributes as source files
    
    font_class = 'fixed'
    
    identifier_specials = '$'

    external_editors = text_editors

    def convert_file(self, path, usage=0):
        if usage == 1:   # indexing
            return extract_source_text(path, '$') ### self.identifier_specials)
        else:
            return extract_plain_text(path)


class Other_Type(Other_Genus):

    def __init__(self, name, attrib, *extensions):
        Other_Genus.__init__(self, name, attrib)
        self.file_extensions = extensions


class Source_Type(Source_Genus):

    def __init__(self, name, specials, *extensions):
        Source_Genus.__init__(self, name)
        self.identifier_specials = specials
        self.file_extensions = extensions


# ---------- Predefined file-type class instances ----------

# Map file_type (from database tuples) into an appropriate
# File_Base (subclass) instance.  Also map file extension
# into a class instance.  These dictionaries are extended by
# file type handlers in directory 'types' via the
# register_type_class procedure.

file_type_class  = {}      # type_name       -> type_class
file_ext_class   = {}      # file_extension  -> type_class
file_ext_name    = {}      # file_extension  -> type_name
genus_species    = {}      # genus_name      -> list of species_names
genus_extensions = {}      # genus_name      -> list of file_extensions

def register_type_class(type_class):
    type_name  = type_class.file_type_name()
    extensions = type_class.file_extensions
    if file_type_class.get(type_name):
        ### add an error or warning dialog???
        emit_log_entry(ops_log_file,
                       "File type class '%s' already exists. "
                       "Second definition will be ignored." % type_name)
    else:
        file_type_class[type_name] = type_class
        for ext in extensions:
            ### check for file extension conflicts???
            file_ext_class[ext] = type_class
            file_ext_name[ext]  = type_name
        genus_species[type_class.genus_name].append(type_class.species_name)
        genus_extensions[type_class.genus_name].extend(extensions)


# Predefined type class instances include the genuses and implicit
# file types (species).

predefined_types = (
    Text_Genus(''),
    Node_Genus(''),
    Exec_Genus(''),
    Other_Genus(''),
    
    Source_Genus(''),
    Document_Genus(''),
    Markup_Genus(''),
    Archive_Genus(''),
    Email_Genus(''),
    
    Directory_Type('direc'),
    Link_Type('link'),
    Binary_Exec_Type('binary'),
    Script_Exec_Type('script'),

    Text_Type('no_ext', ''),
    Text_Type('plain', 'txt'),
    Other_Type('no_ext', 'other'),
    Other_Type('misc', 'other'),
    Other_Type('oversize', 'oversize'),
    )

# All genus classes must be declared in this module.

for type_class in predefined_types:
    genus_species[type_class.genus_name]    = []
    genus_extensions[type_class.genus_name] = []
    
for type_class in predefined_types:
    register_type_class(type_class)


#  ---------- Factory functions and their relatives ----------

def add_source_type(name, specials, *extensions):
    type_class = Source_Type(name, specials, *extensions)
    register_type_class(type_class)

def add_source_extensions(name, *extensions):
    type_name = 'source/%s' % name
    type_class = file_type_class.get(type_name)
    if type_class:
        type_class.file_extensions += extensions
        for ext in extensions:
            ### check for file extension conflicts???
            file_ext_class[ext] = type_class
            file_ext_name[ext]  = type_name
    else:
        ### add an error or warning dialog???
        emit_log_entry(ops_log_file,
                       "File type class '%s' does not exist. "
                       "File extensions will be ignored." % type_name)


#  ---------- Compression attributes ----------

# Each supported compression method is represented by a tuple of the form:
#   (<decompression command>, <file extension 1>, <extension 2>,...)

compression_attributes = \
    { 'gzip' : ('gunzip -c %s', 'gz'),
      'bzip2': ('bunzip2 -c %s', 'bz2', 'bz'),
      }

compress_ext_name    = {}             # file_extension  -> compression name
compress_ext_command = {}             # file_extension  -> compression command

for item in compression_attributes.items():
    for c in item[1][1:]:
        compress_ext_name[c] = item[0]
for attrib in compression_attributes.values():
    for c in attrib[1:]:
        compress_ext_command[c] = attrib[0]
