# Given a user's 4-tuple of search terms, the following functions
# are used to formulate and submit an equivalent query against
# the Pathena database.


import os, sys, time, string, re

from logging_services  import *
from db_connect        import *


#---------------


min_time_values   = ('1970', '1', '1', '0', '0', '0')

# Date-time strings are converted to seconds since the epoch.

def convert_dates(date_bounds):
    if date_bounds.strip() == '': return ()
    def make_mtime(date_str, upper=0):
        vals = date_str.split('/')
        num = len(vals)
#        # bump up low order value for upper bound
#        if upper: vals[len-1] = str(int(vals[len-1]) + 1)
        if num < 6:
            rest = '/'.join(min_time_values[num:])   # missing values
            date_str = '/'.join(vals) + '/' + rest
        time_list = list(time.strptime(date_str, '%Y/%m/%d/%H/%M/%S'))
        time_list[8] = -1       # force DST to be used as needed
        return int(time.mktime(time_list))
    try:
        dates = date_bounds.split()
        from_ = make_mtime(dates[0])
        if len(dates) > 1: till = make_mtime(dates[-1], 1)
        else:              till = int(time.time())
        return from_, till
    except:
        raise SearchTermError, error_messages['bad_dates']

# Parse query string into terms, allowing for quotes, escapes, and
# special terms re'...'.  Returns (plain_terms, regexp_terms, neg_regexp).
### check for embedded quotes or apply escaping ???

def parse_query_terms(query, regexp=0):
    try:
        terms = parse_query_terms_any(query)
    except SearchTermError:
        raise SearchTermError, error_messages['terms_malformed']
    if (terms[1] or terms[2]) and not regexp:
        raise SearchTermError, error_messages['no_regexp_terms']
    return terms
        
def parse_query_terms_any(query):
    query = query.strip()
    parsed_terms = [[], [], []]
    if not query: return parsed_terms
    while query:
        if query[0] in '-!' and len(query) > 3 and \
               query[1:4].lower() in ('re"', "re'"):
            regexp = 2
            query = query[3:]
        elif len(query) > 2 and query[0:3].lower() in ('re"', "re'"):
            regexp = 1
            query = query[2:]
        else:
            regexp = 0
        if query[0] in '\'"':
            start = 1
            while 1:
                # find right end of quoted string,
                # allowing for possible interior quote chars
                end = query.find(query[0], start)
                if end < 0:
                    raise SearchTermError    # missing right quote
                term = query[:end+1]
                try:
                    term = eval(term)
                    break  # clean parse => use it
                except:
                    start = end + 1   # must have escaped quotes; try again
        else:
            start = 0
            while 1:
                # find right end of term,
                # allowing for possible interior escaped blanks
                end = query.find(' ', start)
                if end < 0:
                    try:
                        term = eval('"' + query + '"')
                        end = len(query)
                        break
                    except:
                        raise SearchTermError    # malformed somehow
                elif query[end-1] == '\\':   # must be a previous char
                    # remove backquote and continue scanning
                    query = query[:end-1] + query[end:]
                    start = end   # accounts for shrinkage of one char
                else:
                    try:
                        term = eval('"' + query[:end] + '"')
                        break
                    except:
                        raise SearchTermError    # malformed somehow
        parsed_terms[regexp].append(term)
        query = query[end+1:].lstrip()
        ### check on termination
    return parsed_terms


valid_path_term    = re.compile('(-|!)?\.?\w+')

# NOTE: following check and disjunctive terms not implemented yet.
# A term has one or more disjoined subterms, each of which may be negated
# Each subterm contains non-whitespace characters.

valid_content_term = re.compile('(-|!)?\S+(\|(-|!)?\S+)*')

def valid_search_terms(terms, pattern):
    for t in terms:
        m = pattern.match(t)
        if m and m.group() == t: continue
        return 0
    return 1

def path_name_where(patterns, f_type, *mtimes):
    # Generate a WHERE clause to find path matches based on one or more
    # name fragment patterns.  Successful matches are paths satisfying
    # every one of the fragment patterns.  Include optional subclauses
    # to constrain by file type and mod_time bounds.  Patterns can be
    # either words or regular expressions written in re'...' format.
    # File type is considered a single regular expression string.

    tsearch_terms, regexp_terms, neg_regexps = patterns
    def tsquery_term(term):
        if term[0] in '!-': words = term[1:]
        else:               words = term
        conjunction = encode_compound(words)
        if term[0] in '!-': return '!%s' % conjunction
        return conjunction
    if tsearch_terms:
        terms = [ tsquery_term(t) for t in tsearch_terms ]
        tsearch_clauses = ["path_vector @@ to_tsquery('path_name', '%s')" %
                           '&'.join(terms)]
    else:
        tsearch_clauses = []
    if mtimes:
        time_clauses = ["%d <= mod_time" % mtimes[0],
                        "mod_time <= %d" % mtimes[1]]
    else:
        time_clauses = []
    if not f_type or f_type == any_file_type:
        f_type_clause = []
    elif f_type[0] in '!-':
        f_type_clause = ["file_type !~ '^%s'" % f_type[1:]]  ### left-anchored
    else:
        f_type_clause = ["file_type ~ '^%s'" % f_type]       ### left-anchored

    pos_regexp_clauses = [ "abs_path ~ '%s'" % term for term in regexp_terms ]
    neg_regexp_clauses = [ "abs_path !~ '%s'" % term for term in neg_regexps ]
    subtests = " AND ".join(tsearch_clauses + time_clauses + f_type_clause +
                            pos_regexp_clauses + neg_regexp_clauses)
    if subtests:
        return "WHERE " + subtests
    else:
        return ""

def path_name_query(patterns, f_type, *mtimes):
    return "SELECT file_id, abs_path, file_type, length(content_vector)" \
           " FROM file_descriptors " \
           + path_name_where(patterns, f_type, *mtimes) + \
           " order by mod_time desc"

# Do full-text search using stored procedures.

def full_text_query(terms, *mtimes):
    ### might want to add f_type case here
    if mtimes:
        query = ("select * from find_file_time(%s, %s, %s)",
                 Tsearch_terms(terms), str(mtimes[0]), str(mtimes[1]))
    else:
        query = ("select * from find_file(%s)", Tsearch_terms(terms))
    return query

# If both path and content search is needed, first do path search and
# save results in a temporary table.

def do_conjoined_query(patterns, c_terms, f_type, *mtimes):
    # query on path names first; save in temp table
    do_query_effect(
      "SELECT file_id INTO TEMP phase_1_results FROM file_descriptors "
      + path_name_where(patterns, f_type, *mtimes) )
    results = do_query("select * from find_file_phase2(%s)",
                       Tsearch_terms(c_terms))
    do_query_effect("DROP TABLE phase_1_results")
    return results


# Perform a Pathena query by parsing and validating the user's search terms,
# then formulating suitable SQL queries, and finally, submitting the
# SQL queries to the database.  If supplied, the error_fn accepts two
# string arguments, a title/category and an error message.  If not supplied,
# error messages are directed to sys.stderr.
# Returns a pair of values:
#   - 4-tuple of parsed query strings
#   - list of result triples: [file_id, abs_path, file_type]
# or (None, None) if null or invalid query.

def perform_query(query_strings, error_fn=None):
    # (content terms, path terms, date terms, file type)
    c_query, p_query, d_query, f_query = query_strings
    try:
        mtimes = convert_dates(d_query)
        if c_query:
            c_terms = parse_query_terms(c_query, 0)[0]
            if vacuous_content_query(c_terms):
                raise SearchTermError, error_messages['vacuous_content_query']
# #            if not valid_search_terms(c_terms.split(), valid_content_term):
# #                user_dialog(showerror, *error_messages['bad_content_terms'])
# #                raise SearchTermError
            if p_query or f_query:
                p_terms = parse_query_terms(p_query, 1)
                return ((c_terms, p_terms, mtimes, f_query),
                        do_conjoined_query(p_terms, c_terms, f_query, *mtimes))
            else:
                return ((c_terms, [], mtimes, f_query),
                        do_query(*full_text_query(c_terms, *mtimes)))
        elif p_query or f_query or mtimes:
            p_terms = parse_query_terms(p_query, 1)
            ### currently filtering out numeric terms, but direc names are
            ### allowed to be numeric; consider changing tsearch config to
            ### allow integers and dropping vacuous check below;
            if not filter(None, p_terms[1:]) and p_terms[0] and \
               vacuous_path_query(p_terms[0]):
                raise SearchTermError, error_messages['vacuous_path_query']
            return (([], p_terms, mtimes, f_query),
                    do_query(path_name_query(p_terms, f_query, *mtimes)))
        else:
            return (None, None)
    except SearchTermError, err_tuple:
        if error_fn:
            error_fn(*err_tuple.args)
            return (None, None)
        else:
            print >> sys.stderr, wrap_string('[%s] %s' % err_tuple.args)
            raise SearchTermError


error_messages = \
    {
     'bad_dates' :
     ('Search term processing',
      "The date values have been improperly formed. Use the format "
      "y/m/d/h/m/s (numeric values truncated at any point)." ) ,

     'terms_malformed' :
     ('Search term processing',
      "The search terms are malformed somehow. Check for unbalanced "
      "quotes or misplaced escape characters." ) ,

     'no_regexp_terms' :
     ('Search term processing',
      "Content search terms may not include regular expressions." ) ,

     'bad_path_terms' :
     ('Search term processing',
      "The path search terms have been improperly formed." ) ,

     'bad_content_terms' :
     ('Search term processing',
      "The content search terms have been improperly formed." ) ,

     'vacuous_content_query' :
     ('Search term processing',
      "The content search terms are all 'stop words' or other strings "
      "that do not get indexed. No matches on content can result "
      "from this query." ) ,

     'vacuous_path_query' :
     ('Search term processing',
      "The path search terms are all strings "
      "that do not get indexed. No matches on path name can result "
      "from this query." ) ,
     }
