
# When the client user initiates an indexing operation, the process is
# managed by this module.  The actual indexing is performed by a
# separate process.  Pipes will be established between the client and
# the indexing process to enable control and receive progress updates.
# A listener thread is started to accept information from the pipe.

import os, sys, time, string, glob, re, signal

from app_globals    import *
from client_globals import *
from widget_utils   import *

from file_type_attribs import *

import thread, Queue


#-----------------

end_signals = ('done', 'warning', 'error', 'canceled')

class ProgressState(Queue.Queue):
    def __init__(self, delay):
        Queue.Queue.__init__(self)
        self.state = 'active'
        self.start_time = int(time.time())
        self.cum_time = 0
        self.delay_value = delay

# Handle generic progress update protocol.  Each update is a single text
# line to be interpreted as needed by originator.  End of processing is
# indicated by a value in end_signals.

def listen_for_progress(pipe, queue):
    while 1:
        update = pipe.readline()
        if not update: break    ### other actions?
        update = tuple(update.split())
        queue.put(update)
        if update and update[0] in end_signals: break
    pipe.close()

# Create and pack a slider widget to collect delay value.  Returns the widget.

def delay_slider(parent, init_value=0):
    fr = Frame(parent)
    Label(fr, text='Delay (ms):   ').pack(side=LEFT)
    slider = Scale(fr, length=200, orient=HORIZONTAL, to=200,
                   tickinterval=50)
    slider.set(init_value)
    slider.pack(side=LEFT)
    fr.pack(pady=10)
    return slider

# Background indexing is initiated by first asking the user for his
# profile name, then offering him a chance to set the pause value.
# In the first phase, files matching the profile are identified and
# summary counts are displayed.

def scan_for_files(frame, purge_old=1, update_stats=1):
    profile = askopenfilename(title='Select Indexing Profile',
                              initialdir=profile_dir)
    if not profile:
        frame.destroy(); return
    profile_name = os.path.basename(profile)
    def make_window(fr):
        Message(fr, aspect=250, justify=LEFT, font=text_font,
                text=display_messages['background_indexing']
                ).pack(padx=10, pady=10)
        slider = delay_slider(fr)
        button_row(fr, (('Start Background Indexing',
                         lambda : background_indexing(fr,
                                      profile_name, slider.get(),
                                      purge_old, update_stats), fr),
                        ('Cancel', lambda : fr.destroy(), 0))
                       )[0].pack(pady=10)
    pop_up_toplevel_window(make_window, (), 'Pathena File Indexing',
                           bottom_margin=5)
    frame.destroy()

def build_progress_table(parent):
    rows = (('    ', '     Path', '   Content', '   '),
            ('  Files matching profile:     ', '0', '0', ''),
            ('  Indexed so far:', '0', '0', ''),
            ('  Examined so far:', '0', '0', ''))
    return build_table(parent, (lambda w,t: Label(w, text=t),) * 4,
                       rows, (W, E, E, E))

# Progress update protocol:
#   ('process_id', ...), ('file_counts', ...), ('delete_count', ...),
#   ('current_counts', ...), ..., ('current_counts', ...),
#   '(update_stats yes/no)', 'done'

# The main procedure interface takes the profile name and a few control
# values to start the indexing process.

def background_indexing(prev_fr, profile, delay, purge_old, update_stats):
    prev_fr.destroy()
    progress = ProgressState(delay)
    dynamic_color = 'dark blue'
    success_color = 'green4'
    final_message = {
        'done':     ('completed successfully', success_color),
        'warning':  ('completed with warnings', colors['muted_red']),
        'error':    ('terminated abnormally due to an error', 'red'),
        'canceled': ('been canceled by request', 'blue') }

    # A progress window provides status information and allows the user
    # make changes after indexing has begun.
    
    def make_window(fr):
        def cancel_indexing():
            if user_dialog(askquestion,
                           *dialog_messages['confirm_indexing_cancel']) \
               == 'no':               
                return
            try:
                print >> pipes[0], 'cancel'
                for i in (0,1): pipes[i].close()
            except:
                pass   # process already dead
            progress.state = 'cancel_wait'
            for i in (0,1,2): buttons[i]['state'] = DISABLED
            
        def pause_indexing():
            try:
                print >> pipes[0], 'pause'
                pipes[0].flush()
            except:
                progress.state = 'done'    # process died already
                return
            progress.state = 'paused'
            progress.cum_time += int(time.time()) - progress.start_time
            fr_slider = Frame(fr, relief=SOLID, bd=2)
            Label(fr_slider, text='You may update the per-segment '
                                  'delay value before resuming:'
                  ).pack(padx=5, pady=5)
            progress.delay_slider = delay_slider(fr_slider,
                                                 progress.delay_value)
            fr_slider.pack(before=button_fr, pady=5)
            progress.fr_slider = fr_slider
            buttons[0].configure(state=DISABLED)   # pause
            buttons[1].configure(state=NORMAL)     # resume
            
        def resume_indexing():
            try:
                progress.delay_value = progress.delay_slider.get()
                print >> pipes[0], 'resume', progress.delay_value
                pipes[0].flush()
            except:
                progress.state = 'done'    # process died already
                return
            progress.state = 'active'
            progress.start_time = int(time.time())
            progress.fr_slider.destroy()
            buttons[0].configure(state=NORMAL)     # pause
            buttons[1].configure(state=DISABLED)   # resume

        # A "progress table" shows the file scan counts and displays how
        # many have been completed so far.
        
        Message(fr, aspect=600, justify=LEFT, font=text_font,
                text=display_messages['indexing_in_progress'] % profile
                ).pack(padx=10, pady=10)
        fr2 = Frame(fr, width=200)
        Label(fr2, fg=dynamic_color, text='Indexing time:    ').pack(side=LEFT)
        clock = Label(fr2, fg=dynamic_color, text='0:00:00')
        clock.pack(side=RIGHT)
        fr2.pack(pady=5)
        fr3, prog_table = build_progress_table(fr)
        fr3.configure(relief=SOLID, bd=2)
        fr3.pack(pady=5)
        fr4, prog_msgs = \
             scrolled_text_widget(fr, width=50, height=8, font=small_text_font)
        fr4.pack(pady=10)
        def add_prog_msg(msg, color='black'):
            tag = 'prog_msg_%s' % color
            prog_msgs.insert(END, msg + '\n', tag)
            prog_msgs.tag_configure(tag, foreground=color)
            prog_msgs.see(END)
        add_prog_msg("Scanning for files that match profile '%s'. "
                     "This might take a minute or two..." % profile)
        button_fr, buttons = \
            button_row(fr, (('Pause', pause_indexing),
                            ('Resume', resume_indexing, 0, 1),
                            ('Cancel', cancel_indexing),
                            ('Close', lambda : fr.destroy(), 0, 1)),
                       'Indexing', size=1)
        button_fr.pack(pady=10)
        file_counts = [0, 0]
        indexing_process_ids = []
        
        def indexing_process_done(msg, color='black'):
            add_prog_msg('\nIndexing process has %s.' % msg, color)
            for i in (0,1): pipes[i].close()
            for i in (0,1,2): buttons[i].configure(state=DISABLED)
            buttons[3].configure(state=NORMAL)    # close
            progress.state = 'close_wait'

        # The indexing process can send back progress updates as short
        # text messages having keywords and numbers.  After each phase,
        # progress information is posted to a scrolling display.
        
        def handle_progress_update(update):
            if update[0] in end_signals:
                progress.state = update[0]
                if update[0] in ('done', 'warning'):
                    if update_stats:
                        add_prog_msg("...done.\n", success_color)
                    if update[0] == 'warning':
                        add_prog_msg("NOTE: %s.\n" % ' '.join(update[1:]))
                    if progress.current_counts != tuple(file_counts):
                        add_prog_msg("Some files were not indexed because "
                                     "they already exist in the database, or "
                                     "the profile specifies overlapping sets, "
                                     "or conversion errors occurred.")
                elif update[0] == 'error':
                    add_prog_msg("...aborted.\n", 'red')
                    add_prog_msg("Indexing error: %s." % ' '.join(update[1:]))
                return
            if update[0] == 'current_counts':
                for c in (1,2):
                    prog_table[2][c].configure(text=str(update[c]))
                    prog_table[3][c].configure(text=str(update[c+2]))
                progress.current_counts = int(update[1]), int(update[2])
            elif update[0] == 'update_stats':
                add_prog_msg("...done.\n", success_color)
                if update[1] == 'yes':
                    add_prog_msg("Updating database statistics.  "
                                 "This could take a few minutes...")
                else:
                    add_prog_msg("Updating of database statistics "
                                 "was not selected.\n")
                buttons[0].configure(state=DISABLED)      # pause ###???
            elif update[0] == 'delete_count':
                if update[1] != 'none':
                    add_prog_msg("...done.\n", success_color)
                    add_prog_msg("Purged %s deleted files from database.\n"
                                 % update[1])
                add_prog_msg("Starting file indexing.  "
                             "This could take a while (many minutes)...")
            elif update[0] == 'file_counts':
                add_prog_msg("...done.\n", success_color)
                file_counts[0], file_counts[1] = int(update[1]), int(update[2])
                for c in (0,1):
                    prog_table[1][c+1].configure(text=str(file_counts[c]))
                if purge_old:
                    add_prog_msg("Purging deleted files from database.  "
                                 "This might take another minute or two...")
                else:
                    add_prog_msg("Purging of deleted files was "
                                 "not selected.\n")
            elif update[0] == 'process_id':
                indexing_process_ids.extend(update[1:])

        # Once per second the indexing progress is checked and status
        # information is updated for the user.
        
        def indexing_progress():
            if progress.state == 'active':
                cum_time = progress.cum_time + \
                           int(time.time()) - progress.start_time
                hms = cum_time//3600, (cum_time%3600)//60, cum_time%60
                clock.configure(text='%d:%02d:%02d' % hms)
            elif progress.state in end_signals:
                indexing_process_done(*final_message[progress.state])
                emit_log_entry(main_log_file, "End of indexing process.")
                return
            elif progress.state == 'cancel_wait':
                # give indexer time to react to cancellation request
                progress.state = 'cancel_wait2'
            elif progress.state == 'cancel_wait2':
                # graceful termination failed; kill indexing process
                for pid in indexing_process_ids:
                    try:
                        os.kill(int(pid), signal.SIGKILL)  # force termination
                    except OSError:
                        pass              # process already gone
                # finish log file for indexing process
                indexing_log = \
                    open(os.path.join(log_dir, indexing_log_name), 'a')
                emit_log_entry(indexing_log,
                               '+++++ Indexing canceled by client.')
                emit_log_entry(indexing_log, '<<<<< End of file indexing.')
                indexing_log.close()
                progress.state = 'canceled'
            while 1:
                try:
                    update = progress.get_nowait()
                    handle_progress_update(update)
                except Queue.Empty:
                    break
            fr.update()       # update_idletasks() too weak here
            fr.after(1000, show_progress)

        # Initiate progress tracking.  It will reschedule itself as
        # needed after every pass.
        
        show_progress = EventHandler('Show indexing progress',
                                     indexing_progress, 0)
        emit_log_entry(main_log_file,
                       "Begin indexing process for profile '%s'." % profile)
        pipes = os.popen2('nice -n 15 %s %s %s 1 %s %s 2>> %s' %
                          (os.path.join(bin_dir, 'index_files'),
                           os.path.join(profile_dir, profile), delay,
                           purge_old, update_stats,
                           os.path.join(log_dir, indexing_err_log)))
        fr.after(100, show_progress)
        thread.start_new_thread(listen_for_progress,
                                (pipes[1], progress))
    pop_up_toplevel_window(make_window, (), 'Pathena File Indexing',
                           corner=1, bottom_margin=5)

