# Path and content indexing.  Invoked from a script and runs as a
# separate process.  Optionally can be invoked from the client, where
# it reports progress on standard output while checking for commands
# on standard input.


import os, sys, time, string, glob, shutil
from operator import add, and_
from StringIO import StringIO
from traceback import print_exc

from app_globals       import *
from logging_services  import *
from db_operations     import *

from platform_utilities import *
from file_type_attribs import *

# Normal events and user profile errors are written to the main log.
# Exceptions, database errors and other anomalies are sent to the error log.

main_log_file = open(os.path.join(log_dir, indexing_log_name), 'a')
err_log_file  = open(os.path.join(log_dir, indexing_err_log), 'a')

sys.stderr = err_log_file

### add exception recording and announcement
try:
    for handler in glob.glob(os.path.expandvars('$PATHENA/types/*.py')):
        __import__(os.path.splitext(os.path.basename(handler))[0])
except:
    emit_log_entry(err_log_file, 'Failed to import types modules.')

import exceptions
class CancelProcess(exceptions.Exception):  pass
class BadProfile(exceptions.Exception):  pass

reporting_interval = 1.0  # time between progress reports (sec)
work_interval      = 0.1  # work unit duration (sec)

indexed_file_max_size = 100000000   # limit content indexing to smaller files

_temp_file = new_temp_file_name()
_temp_file_decomp = new_temp_file_name()

see_log_msg = ' See file ~/.pathena/log/%s for details.' % indexing_log_name
see_err_msg = ' See file ~/.pathena/log/%s for details.' % indexing_err_log

do_query_effect("select init_indexing_ops()")

###retain_db_connection()  # Keep database connection open after errors


#--------------------- Command Listening ----------------------

# When invoked from the client, an active connection between client
# and indexer is maintained.  A separate listener thread is used to
# respond to client requests.

import thread, Queue

class CommandState(Queue.Queue):
    def __init__(self, delay):
        Queue.Queue.__init__(self)
        self.delay = delay
        self.delay_sec = delay / 1000.0
        self.break_time = time.time() + work_interval

# Thread function to handle generic control protocol.
# Accept commands from calling/initiating process.
# Each command is a single line of text from stdin.
# Cancellations are accepted at any time.
# Can throw CancelProcess exception.

def listen_for_commands(queue):
    while 1:
        try:
            signal = sys.stdin.readline()
            if not signal: break    # pipe closed
            signal = tuple(signal.split())
            queue.put(signal)
            if signal[0] == 'cancel': break
        except:
            break     # pipe closed by client

# Check for and act on pause and resume commands.
# Only accepted when indexing procedures stop to allow a break.

def accept_pause_cancel(cmd_state):
    active = 1
    while 1:
        try:
            signal = cmd_state.get_nowait()
        except Queue.Empty:
            if active: return
            time.sleep(1.0)    # in pause mode
            continue
        if signal[0] == 'cancel':
            raise CancelProcess
        elif signal[0] == 'pause':
            active = 0
            emit_log_entry(main_log_file, 'Pausing indexing job')
            time.sleep(1.0)
        elif signal[0] == 'resume':
            emit_log_entry(main_log_file, 'Resuming indexing job')
            cmd_state.delay = int(signal[1])
            cmd_state.delay_sec = cmd_state.delay / 1000.0
            return
        else:   # unexpected command
            return

# Check if it's time to take a break.

def check_for_break(cmd_state):
    if time.time() > cmd_state.break_time:
        time.sleep(cmd_state.delay_sec)
        cmd_state.break_time = time.time() + work_interval
#    emit_log_entry(main_log_file, cmd_state.break_time)


#-----------------------  File System Scanning   ------------------------

### should handle multiline directives; also might need to add support
### for quoted patterns.

# Parse profile and return a list [K_1,...,K_n] of directives by keyword.
# K_i is a list of directive lines [D_1,...,D_m] for keyword k_i.
# Each D_j has the leading keyword removed.

def parse_profile(profile_filename, keywords):
    profile_file = open(profile_filename)
    profile_text = profile_file.readlines()
    profile_file.close()
    # strip comments and blank lines, then tokenize
    profile = [ p.split()
                for p in [ line.split('#')[0] for line in profile_text ]
                if p.strip() ]
    directives = [ [ s[1:] for s in profile if s[0] == k ] for k in keywords ]
    bad_keywords = [ s[0] for s in profile if s[0] not in keywords ]
    if filter(None, bad_keywords):
        emit_log_entry(main_log_file,
            "*** Profile error in file '%s'\n"
            "***   Invalid directives: %s" % (profile_filename, bad_keywords))
        raise BadProfile
    return directives

# Apply a generic process of scanning directories for candidate files,
# requesting confirmation, then conducting file indexing.

def scan_by_profile(profile, notify):
    keywords = ('special', 'exclude', 'ignore', 'path', 'content', 'compress')
    try:
        special, exclude, ignore, path, content, compress = \
            parse_profile(profile, keywords)
    except:
        emit_log_entry(main_log_file,
            "*** Error reading or parsing profile '%s'" % profile)
        raise BadProfile
    for c in special + content:
        if len(c) < 2:
            emit_log_entry(main_log_file,
                "*** Profile error: genus or species missing from %s" % c)
            raise BadProfile
    special_lists = [ s[0:2] + expand_paths([s[2:]]) for s in special ]
    exclude_list = expand_paths(exclude)
    target_trees  = expand_paths(path)
    
    reg_types = {}
    # content directives of form: '<genus> <species-pattern-1> ...'
    for g in [ c[0] for c in content ]: reg_types[g] = []
    for c in content:
        species = genus_species[c[0]]
        for p in c[1:]:  # p = species pattern
            rp = re.compile(p+'$')
            reg_types[c[0]].extend([ s for s in species if rp.match(s) ])
#        if '*' in c[1:] : reg_types[c[0]] = genus_species[c[0]]
#        else:             reg_types[c[0]].extend(c[1:])
    try:
        compress_extensions = \
            reduce(add, [ list(compression_attributes[c][1:])
                          for c in reduce(add, compress, []) ], [])
    except KeyError:
        emit_log_entry(main_log_file,
                       "*** Profile error: unknown compression type '%s'" % c)
        raise BadProfile
    return scan_for_files(special_lists, exclude_list, ignore, target_trees,
                          reg_types, compress_extensions, notify)

# Make a pass for each file type to collect info for selected files.
# Collection functions return a triple (type_name, visit_paths, content_paths).

def scan_for_files(specials, exclude_list, ignore_list, target_trees,
                   reg_types, compress_extensions, notify):
    exclude_paths = []
    index_lists = []
    oversize_files = []

    # First collect special files and directories.
    for spec in specials:
        file_type = '%s/%s' % tuple(spec[0:2])
        path, content, oversize = collect_special_files(file_type, *spec[2:])
        index_lists.append((file_type, path, content))
        oversize_files.extend(oversize)
        exclude_paths.extend(spec[2:])
    exclude_paths.extend(exclude_list)
    ignore_patterns = reduce(add, ignore_list, [])

    # Collect implicit and regular files to be indexed by path or content.
    # Path and oversize lists include paths only; content lists include
    # pairs (path, flag), where flag means use path words as well as content.
    
    implicits, extensions = [], []
    for genus, species in reg_types.items():
        for sp in species:
            file_type = '%s/%s' % (genus, sp)
            fclass = file_type_class.get(file_type)
            if file_type in implicit_content_types:
                implicits.append(file_type)
            elif fclass:
                extensions.extend(fclass.file_extensions)
            else:    # bad type  ### collect for reporting?
                emit_log_entry(main_log_file,
                               '*** Erroneous file type: %s' % file_type)
    
    # First collect implicit file/path types.
    index_lists.extend(
        collect_implicit_files(target_trees, exclude_paths, ignore_patterns,
                               implicits, *implicit_file_types))

    path, content, compressed, oversize = \
          collect_regular_files(target_trees, exclude_paths, ignore_patterns,
                                extensions, compress_extensions)
    
    content = remove_duplicate_content(index_lists, content)
    index_lists.append(('*regular*', [], content))
    index_lists.append(('*compressed*', [], compressed))
    
    # Any of these files not indexed by content will be indexed by
    # path only.  If they have recognizable extensions,
    # they will be tagged later with the corresponding types;
    # otherwise, they will be added to genus 'other'.
    
    other_paths = remove_duplicate_paths(index_lists, path)
    index_lists.append(('*regular*', other_paths, []))

    # Collect any oversize files left over from previous passes.
    oversize_files.extend(oversize)
    index_lists.append(('other/oversize', oversize_files, []))
    index_lists = remove_nuisance_files(index_lists)
    path_count, content_count = \
                tuple([ reduce(add, [ len(idx[i]) for idx in index_lists], 0)
                        for i in (1, 2) ])
    if notify:
        # return counts via pipe to calling process
        print 'file_counts', path_count, content_count
        sys.stdout.flush()
    return index_lists

# Expand a list of path name patterns by 'globbing'.  Can raise BadProfile.
# Input is a list of pattern lists.

def expand_paths(plist):
    global warnings
    paths = reduce(add, plist, [])
    epaths = []
    for p in paths:
        eu = os.path.expanduser(p)
        if '*' not in p and '?' not in p and not os.path.exists(eu):
            emit_log_entry(main_log_file,
                           "*** Profile error: path '%s' does not exist." % p)
            raise BadProfile
        path_list = glob.glob(eu)
        if not path_list:
            emit_log_entry(main_log_file,
                "--- Warning: '%s' does not expand into any paths." % p)
            warnings += 1
        epaths.extend(path_list)
    return epaths

# Remove paths that appear on any higher priority lists.

def remove_duplicate_content(index_lists, path_use_pairs):
    prev_paths = {}
    for ftype, visit, content in index_lists:
        for p in visit:   prev_paths[p] = 1
        for p,u in content: prev_paths[p] = 1
    return [ (p,u) for p,u in path_use_pairs if p not in prev_paths ]

def remove_duplicate_paths(index_lists, paths):
    prev_paths = {}
    for ftype, visit, content in index_lists:
        for p in visit:   prev_paths[p] = 1
        for p,u in content: prev_paths[p] = 1
    return [ p for p in paths if p not in prev_paths ]

# Remove nuisance files from index lists.

def keep_file(path):
    return not path.endswith('~')

def remove_nuisance_files(index_lists):
    return [ (ftype, [ p for p in visit if keep_file(p) ],
              [ (p,u) for p,u in content if keep_file(p) ])
             for ftype, visit, content in index_lists ]


# Descend directory trees and collect regular files to be indexed.
# Check each file and subdirectory against the exclude list.
#### collect file length data also???
                          
def collect_regular_files(target_trees, exclude_paths, ignore_list,
                          extensions, compress_extensions):
    visit_paths, content_paths, compressed_paths, oversize = [], [], [], []
    def scan_files(arg, dirpath, namelist):
        if dirpath in exclude_paths:
            del namelist[:]    # don't visit any children
            return
        for a in reduce(add, [ glob.glob1(dirpath, p)
                               for p in ignore_list ], []):
            namelist.remove(a)
        for name in namelist:
            accept_pause_cancel(cmd_state)  # can raise CancelProcess
            check_for_break(cmd_state)
            abs_path = os.path.join(dirpath, name)
            if name.isdigit() and os.path.isfile(abs_path):
                continue          # omit numeric file names
            if abs_path in exclude_paths: continue
            if not os.path.isfile(abs_path): continue
            base, ext = os.path.splitext(name)
            ext = ext[1:]
            if ext in compress_extensions and \
               os.path.splitext(base)[1][1:] in extensions:
                ### check for file size???
                compressed_paths.append((abs_path, 1))
            elif ext not in extensions or \
               ext == '' and not contains_text(abs_path):
                visit_paths.append(abs_path)
            ### need a better way to handle large files:
            elif os.path.getsize(abs_path) < indexed_file_max_size:
                content_paths.append((abs_path, 1))
            else:
                oversize.append(abs_path)
    for tree in target_trees:
        os.path.walk(tree, scan_files, None)
    return visit_paths, content_paths, compressed_paths, oversize

# Descend directory trees and collect special files to be indexed.
# Check each file and subdirectory against the exclude list.
# Special types don't visit nonselected files (their paths are not indexed).

def collect_special_files(special_type, *paths):
    fclass = file_type_class.get(special_type)
    if fclass:
        include_special = fclass.recognize_file
    else:
        emit_log_entry(main_log_file,
                       '*** Erroneous file type: %s' % special_type)
        return [], [], []     # can't continue  ### collect for reporting?
    visit_paths, content_paths = [], []
    def scan_files(arg, dirpath, namelist):
        for name in namelist:
            accept_pause_cancel(cmd_state)  # can raise CancelProcess
            abs_path = os.path.join(dirpath, name)
            use_path, include_content = include_special(abs_path, name)
            if include_content:
                content_paths.append((abs_path, use_path))
            check_for_break(cmd_state)
    for path in paths:
        os.path.walk(path, scan_files, None)
    return visit_paths, content_paths, []

# Descend directory trees and collect implicit files to be indexed.
# Check each file and subdirectory against the exclude list.
#### collect file length data also???
#### assume for now don't have any compressed implicit files
                          
def collect_implicit_files(target_trees, exclude_paths, ignore_list, implicits,
                           dir_name, link_name, bin_name, script_name):
    dir_paths, link_paths, bin_paths, script_paths = [], [], [], []
    dir_content, link_content, script_content = [], [], []
    collecting_direcs  = 'node/direc' in implicits
    collecting_links   = 'node/link' in implicits
    collecting_scripts = 'exec/script' in implicits
    def scan_files(arg, dirpath, namelist):
        if dirpath in exclude_paths:
            del namelist[:]    # don't visit any children
            return
        elif collecting_direcs:
            dir_content.append((dirpath, 1))
        else:
            dir_paths.append(dirpath)
        for a in reduce(add, [ glob.glob1(dirpath, p)
                               for p in ignore_list ], []):
            namelist.remove(a)
        for name in namelist:
            accept_pause_cancel(cmd_state)  # can raise CancelProcess
            check_for_break(cmd_state)
            abs_path = os.path.join(dirpath, name)
            if name.isdigit() and os.path.isfile(abs_path):
                continue          # omit numeric file names
            if abs_path in exclude_paths: continue
            if os.path.islink(abs_path):
                if collecting_links: link_content.append((abs_path, 1))
                else:                link_paths.append(abs_path)
                continue
            if not os.path.isfile(abs_path): continue
            if not os.path.splitext(name)[1] and \
                   os.access(abs_path, os.X_OK):   # found executable
                if contains_text(abs_path):
                    if collecting_scripts: script_content.append((abs_path, 1))
                    else:                  script_paths.append(abs_path)
                else:
                    bin_paths.append(abs_path)
    for tree in target_trees:
        os.path.walk(tree, scan_files, None)
    return ((dir_name, dir_paths, dir_content),
            (link_name, link_paths, link_content),
            (bin_name, bin_paths, []),
            (script_name, script_paths, script_content))


#-------------------  Database access   ------------------

# Index file content using stored procedure.  Ignore if current mod time
# in database is still current.  Return count of files indexed (0 or 1).

def index_file(path, use_path, file_type, decompress_cmd=''):
    global warnings
    try:
        warning_msg = do_transaction(("select index_text_file(%s, %s, %s, %s)",
                                      path, str(use_path),
                                      file_type, decompress_cmd))[0][0]
    except DatabaseError:
        # Connection will have been restarted after error
        do_query_effect("select init_indexing_ops()")
        emit_log_entry(main_log_file,
                       '*** A database error occurred while indexing '
                       'content for file %s.' % path)
        warnings += 1
        index_path(path, file_type)      # try to index path, at least
        return 0   # content not indexed, although path might be

    if warning_msg == '0':
        return 0   # existing tuple is still valid
    elif warning_msg == '1':
        return 0   # file no longer exists (record or report???)
    elif warning_msg:
        emit_log_entry(main_log_file, warning_msg)
        warnings += 1
        return 0   # conversion error or other problem
    return 1

# Index path name only using stored procedure.  Ignore if current mod time
# in database is still current.  Return count of files indexed (0 or 1).

def index_path(path, file_type):
    global warnings
    try:
        cur_mod_time = os.path.getmtime(path)
    except:
        return 0   # file now gone; record/report back???
    db_attribs = do_query("select mod_time, file_type "
                          "from file_descriptors where abs_path = %s", path)
    if db_attribs and \
       db_attribs[0][0] == cur_mod_time and \
       db_attribs[0][1] == file_type:
        return 0       # existing tuple is still valid
    try:
        do_transaction(("select index_path(%s, %s, %s, %s)",
                        path, path_words(path),
                        file_type, str(cur_mod_time)))
    except DatabaseError:
        # Connection restarted after error
        do_query_effect("select init_indexing_ops()")
        warning_msg = '*** A database error occurred while indexing ' \
                      'path %s.' % path
        emit_log_entry(main_log_file, warning_msg)
        warnings += 1
        return 0
    return 1


#----------------------------  Indexing   -----------------------------

# Conduct indexing of files whose paths were collected by a
# previous pass.  Display running count of number completed.
# index_path and index_file operate on absolute path names.

def conduct_indexing(index_lists, notify, purge_old, update_stats):
    global current_file_path
    if purge_old:
        delete_count = do_transaction(("select purge_deleted_files()",))[0][0]
        emit_log_entry(main_log_file,
                       '>>> Purged %d deleted paths from database.' %
                       delete_count)
    else:
        delete_count = 'none'
    if notify:
        print 'delete_count', delete_count
        sys.stdout.flush()
    path_count, content_count = 0, 0
    path_tries, content_tries = 0, 0
    report_time = 0.0
    
    for file_type, visit_paths, content_paths in index_lists:
        # following can throw exceptions
        for path in visit_paths:
            current_file_path = path
            if notify:
                accept_pause_cancel(cmd_state)      # can raise CancelProcess
            if file_type not in ('*regular*',):
                file_type_name = file_type
            else:
                file_ext = os.path.splitext(path)[1][1:]
                if file_ext:
                    file_type_name = file_ext_name.get(file_ext, 'other/misc')
                else:
                    file_type_name = 'other/no_ext'
            path_count += index_path(path, file_type_name)
            path_tries += 1
            if notify:
                cur_time = time.time()
                if cur_time > report_time:
                    print 'current_counts', \
                          path_count, content_count, path_tries, content_tries
                    sys.stdout.flush()
                    report_time = cur_time + reporting_interval
            check_for_break(cmd_state)

        for path, use_path in content_paths:
            current_file_path = path
            if notify:
                accept_pause_cancel(cmd_state)      # can raise CancelProcess
            decompress_cmd = ''
            if file_type not in ('*regular*', '*compressed*'):
                file_type_name = file_type
            else:
                base, ext = os.path.splitext(path)
                ext = ext[1:]
                if file_type == '*compressed*':
                    decompress_cmd = compress_ext_command[ext]
                    file_ext = os.path.splitext(base)[1][1:]    # inner ext
                else:
                    file_ext = ext
                if file_ext:
                    file_type_name = file_ext_name.get(file_ext, 'other/misc')
                else:
                    file_type_name = 'text/no_ext'
            content_count += \
                index_file(path, use_path, file_type_name, decompress_cmd)
            content_tries += 1
            if notify:
                cur_time = time.time()
                if cur_time > report_time:
                    print 'current_counts', \
                          path_count, content_count, path_tries, content_tries
                    sys.stdout.flush()
                    report_time = cur_time + reporting_interval
            check_for_break(cmd_state)

    if notify:
        print 'current_counts', \
              path_count, content_count, path_tries, content_tries
        if update_stats: print 'update_stats yes'
        else:            print 'update_stats no'
        sys.stdout.flush()
    update_summary_stats()
    if update_stats:
        refreshTsearchStats()
    do_query_effect('COMMIT')
    if notify:
        vacuumDatabase(5)
        if warnings:
            print ('warning indexing generated %d warning(s).' % warnings) \
                  + see_log_msg
            emit_log_entry(main_log_file,
                           '*** %d warnings/errors were issued.' % warnings)
        else:
            print 'done'       # signal successful end of job
        sys.stdout.flush()
    emit_log_entry(main_log_file,
                   '<<<<< End of file indexing.\nFiles indexed/tried by '
                   'path, content:  %d / %d, %d / %d' %
                   (path_count, path_tries, content_count, content_tries))

# Progress update protocol/sequence:
#   ('process_id', ...), ('file_counts', ...), ('delete_count', ...),
#   ('current_counts', ...), ..., ('current_counts', ...),
#   '(update_stats yes/no)', 'done'


#-----------------------------------------

current_file_path = ''      # names file if exception raised

# This is the top-level function for externally invoking the indexing
# process.  Set notify=1 to interact via pipes.

def scan_and_index_files(profile, delay=0, notify=0,
                         purge_old=1, update_stats=1):  # for testing
    global cmd_state, warnings
    warnings = 0
    delay, notify = int(delay), int(notify)
    purge_old, update_stats = int(purge_old), int(update_stats)
    if update_stats: index_type = 'full'
    else:            index_type = 'quick'
    print >> main_log_file, ''
    emit_log_entry(main_log_file,
                   ">>>>> Begin %s file indexing for profile '%s'." %
                   (index_type, os.path.basename(profile)))
    trace = StringIO()
    exit_code = 0
    
    try:
        cmd_state = CommandState(delay)
        if notify:
            print 'process_id', os.getpid()
            sys.stdout.flush()
        if notify:
            thread.start_new_thread(listen_for_commands, (cmd_state,))
        index_lists = scan_by_profile(profile, notify)
        conduct_indexing(index_lists, notify, purge_old, update_stats)
        
    except CancelProcess:
        emit_log_entry(main_log_file, '+++++ Indexing canceled by client.')
        emit_log_entry(main_log_file, '<<<<< End of file indexing.')
        if notify:
            print 'canceled'     # signal cancellation to caller
            sys.stdout.flush()
        vacuumDatabase(5)
        exit_code = 1
        
    except BadProfile:
        emit_log_entry(main_log_file, '***** Indexing aborted after '
                       'error in user profile detected.')
        emit_log_entry(main_log_file, '<<<<< End of file indexing.')
        if notify:
            print 'error indexing profile has one or more errors' + see_log_msg
            sys.stdout.flush()
        exit_code = 2
        
    except IOError, exc_obj:
        emit_log_entry(main_log_file,
                       '***** IOError signaled: %s' %
                       ' ; '.join([ str(a) for a in exc_obj.args ]))
        emit_log_entry(main_log_file, '+++++ Indexing aborted.')
        emit_log_entry(main_log_file, '<<<<< End of file indexing.')
        # Don't try to notify because error was most likely a broken pipe.
        vacuumDatabase(5)
        exit_code = 3
        
    except Exception, exc_obj:
        emit_log_entry(main_log_file,
                       '*** Unhandled exception: %s. ' + see_err_msg)
        emit_log_entry(err_log_file,
                       '*** Unhandled exception: %s' %
                       ' ; '.join([ str(a) for a in exc_obj.args ]))
        print_exc(20, trace)
        emit_log_entry(err_log_file, trace.getvalue())
        file_msg = '*** File being indexed when exception occurred:\n   %s' \
                   % current_file_path
        emit_log_entry(main_log_file, file_msg)
        emit_log_entry(err_log_file, file_msg)
        emit_log_entry(main_log_file, '<<<<< End of file indexing.')
        if notify:
            print 'error internal error in indexing process' + see_log_msg
            sys.stdout.flush()
        vacuumDatabase(5)
        exit_code = 4

    try:
        main_log_file.close()
        err_log_file.close()
        os.remove(_temp_file)
        os.remove(_temp_file_decomp)
    except:
        pass
    return exit_code
