--
-- Schema for Pathena: a full-text, desktop search engine.
--

-- Create new Tsearch configurations that drop most tokens from the
-- default configuration, leaving only basic words.

-- Configuration for path name components (no stemming or stop words):

--INSERT INTO pg_ts_cfg VALUES ('path_name', 'default', 'C');
INSERT INTO pg_ts_cfg VALUES ('path_name', 'default', 'en_US.UTF-8');

COPY pg_ts_cfgmap FROM stdin;
path_name	lword	{simple}
path_name	nlword	{simple}
path_name	word	{simple}
path_name	part_hword	{simple}
path_name	nlpart_hword	{simple}
path_name	lpart_hword	{simple}
\.

-- Following were dropped:

--path_name	email	{simple}
--path_name	url	{simple}
--path_name	host	{simple}
--path_name	sfloat	{simple}
--path_name	version	{simple}
--path_name	hword	{simple}
--path_name	lhword	{en_stem}
--path_name	nlhword	{simple}
--path_name	uri	{simple}
--path_name	file	{simple}
--path_name	float	{simple}
--path_name	int	{simple}
--path_name	uint	{simple}


-- Configuration for file content words (with stemming and stop words):

--INSERT INTO pg_ts_cfg VALUES ('file_content', 'default', 'C');
INSERT INTO pg_ts_cfg VALUES ('file_content', 'default', 'en_US.UTF-8');

COPY pg_ts_cfgmap FROM stdin;
file_content	lword	{en_stem}
file_content	nlword	{simple}
file_content	word	{simple}
file_content	part_hword	{simple}
file_content	nlpart_hword	{simple}
file_content	lpart_hword	{en_stem}
\.

-- Following were dropped:

--file_content	email	{simple}
--file_content	url	{simple}
--file_content	host	{simple}
--file_content	sfloat	{simple}
--file_content	version	{simple}
--file_content	hword	{simple}
--file_content	lhword	{en_stem}
--file_content	nlhword	{simple}
--file_content	uri	{simple}
--file_content	file	{simple}
--file_content	float	{simple}
--file_content	int	{simple}
--file_content	uint	{simple}

SELECT set_curcfg('file_content');


-- The main table describing files is file_descriptors, which is
-- heavily indexed using both B-tree and tsearch2 indexes.
-- A subset of files will have their contents (converted to text)
-- partially cached in this table in columns file_title,
-- file_metadata and text_content.  When no caching is used (most
-- of the time), these columns are NULL.
--
-- Two tsvectors are kept: path_vector describes path 'words', and
-- content_vector describes a file's textual content (up to some limit).
-- During file indexing, these vectors will be built using the tsearch2
-- configurations path_name and file_content, respectively.

CREATE TABLE file_descriptors (
    file_id serial PRIMARY KEY,
--    file_id serial NOT NULL UNIQUE,
    abs_path text NOT NULL UNIQUE,
    file_type varchar(20),
    mod_time integer,
    file_title text,
    file_metadata text,
    text_content text,
    path_vector tsvector,
    content_vector tsvector
);


-- This tiny table simply saves summary statistics about file_descriptors
-- to avoid counting when database status is requested.

CREATE TABLE summary_stats (
    num_paths integer,
    num_indexed integer,
    num_cached integer
);

INSERT INTO summary_stats VALUES (0, 0, 0);


-- Search queries return tuples of this type.

CREATE TYPE find_file_t AS (
	file_id integer,
	abs_path text,
	file_type varchar(20),
	rank real
);
--	mod_time integer,


-- An index on column file_type supports regular expression search.

CREATE INDEX file_type_index ON file_descriptors
       (file_type varchar_pattern_ops);


-- The index on column mod_time is conventional.

CREATE INDEX mod_time_index ON file_descriptors (mod_time);


-- A GiST index on path_vector uses what tsearch2 builds.

CREATE INDEX path_name_index ON file_descriptors
USING gist (path_vector);


-- A GiST index on content_vector uses what tsearch2 builds.

CREATE INDEX file_content_index ON file_descriptors
USING gist (content_vector);


-- Histograms of tsearch2 statistics are kept in tables.

CREATE TYPE gist_stats_t AS (
	word text,
	ndoc integer,
	nentry integer
);


-- Create the initial GiST statistics tables.

SELECT * INTO latest_path_stats
FROM stat('select path_vector from file_descriptors');

SELECT * INTO latest_content_stats
FROM stat('select content_vector from file_descriptors');
