--
-- Stored procedures for Pathena
--
--   Assumes basic schema already created.
--   Written in three server-side languages: sql, plpgsql and plpythonu.
--

--
-- init_python_setup(base_dir, pid)
-- Initialize global data dictionary GD for Python functions.
--
-- Note: Due to a Postgres restriction, Python functions cannot
-- return type void, so integer is used instead.
--

CREATE OR REPLACE FUNCTION init_plpythonu_setup(text, integer)
RETURNS integer AS '
    base_dir, pid = args
    if GD.get("initialized"): return
    GD["PATHENA"] = base_dir
    import sys, os
    GD["sys"] = sys
    GD["os"] = os
    for sub_dir in ("lib", "server", "types"):
        sys.path.insert(0, os.path.join(base_dir, sub_dir))
    GD["plpy.execute"] = plpy.execute
    import server_globals
    server_globals.start_up(GD)
    GD["initialized"] = 1
    GD["emit_plpy_log_entry"]("Python GD initialized, process id: %s" % pid)
' language plpythonu;

--    sys.path.insert(0, os.path.join(base_dir, "lib"))
--    sys.path.insert(0, os.path.join(base_dir, "server"))
--    sys.path.insert(0, os.path.join(base_dir, "types"))

-- should be invoked after connection to database server is established:
--   select init_plpythonu_setup(base_dir, client_pid);


-------------------------------------------------------
--
-- index_path(abs_path, path_words, file_type, mod_time)
-- Index path only; file contents are not indexed.
--

CREATE OR REPLACE FUNCTION index_path(text, text, text, integer)
RETURNS void AS '
DECLARE
    path ALIAS FOR $1;
    path_words ALIAS FOR $2;
    ftype ALIAS FOR $3;
    mtime ALIAS FOR $4;
    cur_id INTEGER;
BEGIN
    SELECT INTO cur_id file_id FROM file_descriptors WHERE path = abs_path;
    IF FOUND THEN
       UPDATE file_descriptors
       SET file_type=ftype, mod_time=mtime,
           file_title=NULL, file_metadata=NULL, text_content=NULL,
           path_vector=strip(to_tsvector(\'path_name\', path_words)),
           content_vector=NULL
       WHERE file_id = cur_id;
    ELSE
       INSERT INTO file_descriptors (abs_path, file_type, mod_time, 
                                     file_title, file_metadata, text_content,
                                     path_vector, content_vector)
              VALUES (path, ftype, mtime, NULL, NULL, NULL,
                      strip(to_tsvector(\'path_name\', path_words)), NULL);
    END IF;
    RETURN;
END;
' LANGUAGE plpgsql;


--
-- index_file(abs_path, path_words, file_type, file_text, mod_time)
-- File content (using conversion to text) is indexed (along with path).
--

CREATE OR REPLACE FUNCTION index_file(text, text, text, text, integer)
RETURNS void AS '
DECLARE
    path ALIAS FOR $1;
    path_words ALIAS FOR $2;
    ftype ALIAS FOR $3;
    ftext ALIAS FOR $4;
    mtime ALIAS FOR $5;
    cur_id INTEGER;
BEGIN
    SELECT INTO cur_id file_id FROM file_descriptors WHERE path = abs_path;
    IF FOUND THEN
       UPDATE file_descriptors
       SET file_type=ftype, mod_time=mtime,
           file_title=NULL, file_metadata=NULL, text_content=NULL,
           path_vector=strip(to_tsvector(\'path_name\', path_words)),
           content_vector=to_tsvector(\'file_content\', ftext)
       WHERE file_id = cur_id;
    ELSE
       INSERT INTO file_descriptors (abs_path, file_type, mod_time, 
                                     file_title, file_metadata, text_content,
                                     path_vector, content_vector)
              VALUES (path, ftype, mtime, NULL, NULL, NULL,
                      strip(to_tsvector(\'path_name\', path_words)),
                      to_tsvector(\'file_content\', ftext));
    END IF;
    RETURN;
END;
' LANGUAGE plpgsql;


--
-- index_file_cache(abs_path, path_words, file_type,
--                  title, metadata, cache_text, file_text, mod_time)
-- File contents, path are indexed and text conversion is cached.
--

CREATE OR REPLACE FUNCTION index_file_cache(
                             text, text, text, text, text, text, text, integer)
RETURNS void AS '
DECLARE
    fpath ALIAS FOR $1;
    path_words ALIAS FOR $2;
    ftype ALIAS FOR $3;
    ftitle ALIAS FOR $4;
    fmetadata ALIAS FOR $5;
    cache_text ALIAS FOR $6;
    ftext ALIAS FOR $7;
    mtime ALIAS FOR $8;
    cur_id INTEGER;
BEGIN
    SELECT INTO cur_id file_id FROM file_descriptors WHERE fpath = abs_path;
    IF FOUND THEN
       UPDATE file_descriptors
       SET file_type=ftype, mod_time=mtime,
           file_title=ftitle, file_metadata=fmetadata, text_content=cache_text,
           path_vector=strip(to_tsvector(\'path_name\', path_words)),
           content_vector=to_tsvector(\'file_content\', ftext)
       WHERE file_id = cur_id;
    ELSE
       INSERT INTO file_descriptors (abs_path, file_type, mod_time, 
                                     file_title, file_metadata, text_content,
                                     path_vector, content_vector)
              VALUES (fpath, ftype, mtime, ftitle, fmetadata, cache_text,
                      strip(to_tsvector(\'path_name\', path_words)),
                      to_tsvector(\'file_content\', ftext));
    END IF;
    RETURN;
END;
' LANGUAGE plpgsql;


-------------------------------------------------------
--
-- init_indexing_ops()
-- Import and initialize support operations for indexing.
-- Invoke initialization before starting to index files.
--

CREATE OR REPLACE FUNCTION init_indexing_ops() RETURNS integer AS '
    import indexing_ops
    indexing_ops.init_support_ops(GD)
    GD["emit_plpy_log_entry"]("Indexing ops initialized")
' language plpythonu;

--
-- index_text_file(path, use_path, file_type, decompress_cmd)
-- Contents of text file are indexed; used by indexing process.
--

CREATE OR REPLACE FUNCTION index_text_file(text, integer, text, text)
RETURNS text AS '
    return GD["index_file_text"](*args)
' language plpythonu;


-------------------------------------------------------
--
-- init_client_ops()
-- Import and initialize support for client operations.
-- Invoke initialization before starting up client.
--

CREATE OR REPLACE FUNCTION init_client_ops() RETURNS integer AS '
    import client_ops
    client_ops.init_support_ops(GD)
    GD["emit_plpy_log_entry"]("Client ops initialized")
' language plpythonu;

--
-- get_cached_content(type_name, path, usage)
--

CREATE OR REPLACE FUNCTION get_cached_content(text, text, integer)
RETURNS text AS '
    type_name, path, usage = args
    type_class = GD["file_type_attribs"].file_type_class[type_name]
    results = GD["get_cached_content"](type_class, path, usage)
    return results
' language plpythonu;

--
-- get_headline(type_name, path, terms, limit, specials, mode)
-- Ask Tsearch2 to derive a headline string given query terms.
--

CREATE OR REPLACE FUNCTION get_headline(
                               text, text, text, integer, text, integer)
RETURNS text AS '
    type_name, path, terms, limit, specials, mode = args
    type_class = GD["file_type_attribs"].file_type_class[type_name]
    results = GD["get_headline"](type_class, path, terms,
                                 limit, specials, mode)
    return results
' language plpythonu;


-------------------------------------------------------
--
-- find_file(query_string)
-- Conduct full text query on content terms.
--

CREATE OR REPLACE FUNCTION find_file(text) RETURNS SETOF find_file_t
    AS 'select file_id, abs_path, file_type, rank_cd(content_vector, q)
	from file_descriptors, to_tsquery(\'file_content\', $1) as q
	where content_vector @@ q order by rank_cd(content_vector, q) desc'
    LANGUAGE sql;

--
-- find_file_time(query_string, mtime_low, mtime_high)
-- Conduct query on content terms with mod time constraints.
--

CREATE OR REPLACE FUNCTION find_file_time(text, integer, integer)
    RETURNS SETOF find_file_t
    AS 'select file_id, abs_path, file_type, rank_cd(content_vector, q)
	from file_descriptors, to_tsquery(\'file_content\', $1) as q
	where content_vector @@ q and $2 <= mod_time and mod_time <= $3
        order by rank_cd(content_vector, q) desc'
    LANGUAGE sql;

-- Create temp table so next function can be defined.

SELECT * INTO TEMP phase_1_results
    FROM (select F.file_id from file_descriptors F) AS T;

--
-- find_file_phase2(query_string)
-- Assumes temp table 'phase_1_results' exists (file_ids only).
-- Conduct content search as second step of a two-phase query by
-- joining with file_ids produced in first step.
--

CREATE OR REPLACE FUNCTION find_file_phase2(text) RETURNS SETOF find_file_t
    AS 'select f.file_id, f.abs_path, f.file_type, rank_cd(content_vector, q)
	from file_descriptors f join phase_1_results r
             on (f.file_id = r.file_id),
             to_tsquery(\'file_content\', $1) as q
	where content_vector @@ q order by rank_cd(content_vector, q) desc '
    LANGUAGE sql;

--	from file_descriptors f join
--        (select * from phase_1_results) as r


-------------------------------------------------------
--
-- update_path_stats()
-- Compute word statistics for ts_vector index (path).
--

CREATE OR REPLACE FUNCTION update_path_stats() RETURNS void
    AS 'insert into latest_path_stats
        select * from stat(\'select path_vector from file_descriptors\')
        order by nentry desc'
    LANGUAGE sql;

--
-- update_content_stats()
-- Compute word statistics for ts_vector index (content).
--

CREATE OR REPLACE FUNCTION update_content_stats() RETURNS void
    AS 'insert into latest_content_stats
        select * from stat(\'select content_vector from file_descriptors\')
        order by nentry desc'
    LANGUAGE sql;


-------------------------------------------------------
--
-- purge_deleted_files()
-- Scan indexed files and delete those no longer found in the file system.
--

CREATE OR REPLACE FUNCTION purge_deleted_files() RETURNS integer AS '
    path_exists = GD["os"].path.exists
    del_count = 0
    all_rows = \\
        GD["execute_query"]("select file_id, abs_path from file_descriptors")
    if all_rows == None:
        return 0     ### add failure indication???
    for row in all_rows:
        if not path_exists(row["abs_path"]):
            GD["execute_query"](
                "delete from file_descriptors where file_id = %s" %
                row["file_id"])
            del_count += 1     ### check for failure???
    return del_count
' language plpythonu;


------------------------------------
